@@ -33,6 +33,9 @@
 #define likely(x)       __builtin_expect(!!(x), 1)
 #define unlikely(x)     __builtin_expect(!!(x), 0)
 
+/* pointer alignment test */
+#define IS_PTR_ALIGNED(p, a) (!!(((long int)p & ((a)-1)) == 0))
+
 /*
  * Microcode update file data structures
  */
@@ -272,36 +275,67 @@ static inline int is_valid_bcd(const uint32_t data)
 	       ((((data + 0x06666666U) ^ data) & 0x11111110U) == 0);
 }
 
-/**
- * intel_ucode_check_microcode() - perform sanity checks on a microcode entry
+static inline int is_zero_checksum(const uint8_t *data, uint32_t dwords)
+{
+	uint32_t s = 0;
+
+	/*
+	 * Test using -tr loader on a valid microcode data file with one to
+	 * three bytes of zeros prepended (to offset the valid microcode
+	 * data out of alignment).
+	 */
+
+	if (IS_PTR_ALIGNED(data, sizeof(uint32_t))) {
+		/*
+		 * gcc might generate vectorized code that cannot handle
+		 * unaligned pointer derreferences when given this pattern:
+		 */
+		const uint32_t *p = (uint32_t *)data;
+		while (dwords--)
+			s += *(p++);
+	} else {
+		/*
+		 * Avoid unaligned accesses.  gcc seems to always vectorize
+		 * this properly into code that can handle unaligned data.
+		 * When not vectorized, generates slow code.
+		 *
+		 * Note:
+		 * s += (((uint16_t *)data)[0] | (((uint16_t *)data)[1] << 16));
+		 *
+		 * also works when vectorized by gcc, and generates faster
+		 * and better code than doing it one byte at a time like
+		 * below.  However, when working around compiler issues, it
+		 * often pays off to go all the way...
+		 */
+		while (dwords--) {
+			s += (data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24));
+			data += 4;
+		}
+	}
+
+	return (s == 0);
+}
+
+/*
+ * xx_intel_ucode_check_uc()
  *
- * @uc:		pointer to the beginning of a microcode entry
- * @strict:	if non-zero, perform more strict checking
+ * refer to intel_ucode_check_microcode() for details.
  *
- * This function checks the well-formedness and sanity of a microcode entry.
- * All other functions in the library expect to receive sane and well-formed
- * microcode headers and full microcode entries, so this function MUST be
- * used beforehand.
+ * @uc does not have to be aligned.
  *
- * In strict mode, secondary checks such as size constraints are applied
- * which will help weed off almost-correct data.  This DID flag some weird
- * microcode for signature 0x106c0, present in one of the microcode files
- * distributed by urbanmyth.org in 2008 (apparently it was microcode for
- * an engineering stepping of an Atom processor).
- *
- * Returns INTEL_UCODE_NOERROR if the microcode entry looks sane, or a
- * different status code indicating a problem with the microcode entry.
+ * We depend on the fact that x86/x86-64 is fine with misaligned data
+ * access, as long as you keep away from aligned vector instrictions and
+ * other "specials".
  */
-intel_ucode_status_t intel_ucode_check_microcode(const void * const uc, int strict)
+static intel_ucode_status_t xx_intel_ucode_check_uc(const void * const uc,
+						size_t maxlen, int strict)
 {
 	unsigned long int total_size, data_size, ext_table_size;
 	const struct intel_ucode_v1_hdr * const uc_header = uc;
-	uint32_t sum, orig_sum;
 	unsigned int i;
-	uint32_t *p;
 
-	if (unlikely(!uc))
-		return INTEL_UCODE_BAD_PARAMETERS;
+	if (unlikely(maxlen && maxlen < 1024))
+		return INTEL_UCODE_INVALID_DATA;
 
 	if (unlikely(uc_header->hdrver != 1))
 		return INTEL_UCODE_UNKNOWN_FORMAT;
@@ -311,57 +345,45 @@ intel_ucode_status_t intel_ucode_check_microcode(const void * const uc, int stri
 	total_size = intel_ucode_v1_get_totalsize(uc_header);
 	data_size = intel_ucode_v1_get_datasize(uc_header);
 
-	if (unlikely(data_size > total_size || data_size < INTEL_UC_V1_HEADER_SIZE))
+	if (data_size > total_size || data_size < INTEL_UC_V1_HEADER_SIZE)
 		return INTEL_UCODE_INVALID_DATA;
-	if (unlikely(data_size + INTEL_UC_V1_HEADER_SIZE > total_size))
+	if (data_size + INTEL_UC_V1_HEADER_SIZE > total_size)
 		return INTEL_UCODE_INVALID_DATA;
 
-	if (unlikely(data_size % sizeof(uint32_t)))
+	if (data_size % sizeof(uint32_t))
 		return INTEL_UCODE_INVALID_DATA;
-	if (unlikely(total_size % sizeof(uint32_t)))
+	if (total_size % sizeof(uint32_t))
 		return INTEL_UCODE_INVALID_DATA;
-	if (unlikely(strict && total_size % 1024))
+	if (strict && total_size % 1024)
+		return INTEL_UCODE_INVALID_DATA;
+	if (maxlen && total_size > maxlen)
 		return INTEL_UCODE_INVALID_DATA;
 
 	/* Calculate the checksum.  We exclude the extended table as it
 	 * also has to have a zero checksum, in order to get better
 	 * coverage */
-	orig_sum = 0;
-	i = (INTEL_UC_V1_HEADER_SIZE + data_size) / sizeof(uint32_t);
-	p = (uint32_t *)uc;
-	while (i--) {
-		orig_sum  += *p;
-		p++;
-	}
-	if (unlikely(orig_sum != 0))
+	if (!is_zero_checksum(uc, (INTEL_UC_V1_HEADER_SIZE + data_size) / sizeof(uint32_t)))
 		return INTEL_UCODE_BAD_CHECKSUM; /* invalid checksum */
 
+	/* we can now assume that this is very likely to be microcode */
+
 	ext_table_size = total_size - (INTEL_UC_V1_HEADER_SIZE + data_size);
-	if (ext_table_size) {
+	if (unlikely(ext_table_size)) {
 		const struct intel_ucode_v1_extsig_table *ext_header;
 		const struct intel_ucode_v1_extsig *ext_sig;
 		uint32_t ext_sigcount;
-		uint32_t ext_table_sum;
 
-		if (unlikely(ext_table_size < INTEL_UC_V1_EXTHDR_SIZE))
+		if (ext_table_size < INTEL_UC_V1_EXTHDR_SIZE)
 			return INTEL_UCODE_BAD_EXTENDED_TABLE; /* exttable size too small */
-		if (unlikely((ext_table_size - INTEL_UC_V1_EXTHDR_SIZE) % INTEL_UC_V1_EXTSIG_SIZE))
+		if ((ext_table_size - INTEL_UC_V1_EXTHDR_SIZE) % INTEL_UC_V1_EXTSIG_SIZE)
 			return INTEL_UCODE_BAD_EXTENDED_TABLE; /* bad exttable size */
 
 		ext_header = (void *)((uint8_t *)uc + INTEL_UC_V1_HEADER_SIZE + data_size);
-		if (unlikely(ext_table_size != intel_uc_v1_exttable_size(ext_header)))
+		if (ext_table_size != intel_uc_v1_exttable_size(ext_header))
 			return INTEL_UCODE_BAD_EXTENDED_TABLE; /* bad exttable size */
 
 		/* extended table checksum */
-		ext_table_sum = 0;
-
-		i = ext_table_size / sizeof(uint32_t);
-		p = (uint32_t *)ext_header;
-		while (i--) {
-			ext_table_sum += *p;
-			p++;
-		}
-		if (unlikely(ext_table_sum != 0))
+		if (!is_zero_checksum((uint8_t *)ext_header, ext_table_size / sizeof(uint32_t)))
 			return INTEL_UCODE_BAD_EXTENDED_TABLE_CHECKSUM; /* invalid checksum */
 
 		ext_sigcount = ext_header->count;
@@ -370,9 +392,8 @@ intel_ucode_status_t intel_ucode_check_microcode(const void * const uc, int stri
 		ext_sig = (void *)((uint8_t *)ext_header + INTEL_UC_V1_EXTHDR_SIZE);
 		i = ext_sigcount;
 		while (i--) {
-			sum = orig_sum
-				- (uc_header->sig + uc_header->pf_mask + uc_header->cksum)
-				+ (ext_sig->sig + ext_sig->pf_mask + ext_sig->cksum);
+			uint32_t sum = (ext_sig->sig + ext_sig->pf_mask + ext_sig->cksum) -
+			               (uc_header->sig + uc_header->pf_mask + uc_header->cksum);
 			if (sum)
 				return INTEL_UCODE_BAD_EXTENDED_SIG_CHECKSUM; /* invalid checksum */
 			ext_sig++;
@@ -395,6 +416,133 @@ intel_ucode_status_t intel_ucode_check_microcode(const void * const uc, int stri
 	return INTEL_UCODE_NOERROR;
 }
 
+/**
+ * intel_ucode_check_microcode() - perform sanity checks on a microcode entry
+ *
+ * @uc:		pointer to the beginning of a microcode entry
+ * @strict:	if non-zero, perform more strict checking
+ * @maxlen:	if non-zero, limit microcode entry to maxlen bytes
+ *
+ * This function checks the well-formedness and sanity of a microcode entry.
+ * All other functions in the library expect to receive sane and well-formed
+ * microcode headers and full microcode entries, so this function MUST be
+ * used beforehand.
+ *
+ * @uc must be correctly aligned to a 4-byte boundary.
+ *
+ * It is safe to specify @maxlen as zero if the microcode entry has been
+ * processed by intel_ucode_foreach_microcode().
+ *
+ * In strict mode, secondary checks such as size constraints are applied
+ * which will help weed off almost-correct data.  This DID flag some weird
+ * microcode for signature 0x106c0, present in one of the microcode files
+ * distributed by urbanmyth.org in 2008 (apparently it was microcode for
+ * an engineering stepping of an Atom processor).
+ *
+ * Returns INTEL_UCODE_NOERROR if the microcode entry looks sane, or a
+ * different status code indicating a problem with the microcode entry.
+ */
+intel_ucode_status_t intel_ucode_check_microcode(const void * const uc,
+						 size_t maxlen, int strict)
+{
+	if (unlikely(!uc || !IS_PTR_ALIGNED(uc, sizeof(uint32_t))))
+		return INTEL_UCODE_BAD_PARAMETERS;
+
+	return xx_intel_ucode_check_uc(uc, maxlen, strict);
+}
+
+/**
+ * intel_ucode_scan_for_microcode() - scan for valid microcode
+ *
+ * @bs: 	pointer to a pointer to the memory area to scan for valid
+ * 		microcode.  Modified on return to point to the start of the
+ * 		area with valid microcode
+ * @be: 	Modified on return to point to the first byte past the end
+ * 		of the area with valid microcode
+ * @blen:	Modified on return to be the size of the microcode area
+ * @alen:	size of the memory area to search.  Modified on return
+ *		to account for the bytes that have been already searched
+ *		(i.e. @bs delta plus @blen).
+ *
+ * This function searches for the first well-formed microcode entry in the
+ * memory area starting at @bs, with a length of @alen.  Then, it validates
+ * all microcodes in that memory area that are stored back-to-back, and
+ * updates @bs to point to the start of the first microcode.  It sets @be
+ * to point to the byte after the last microcode and @blen to the size of
+ * the microcode block, and updates @alen to account for all data already
+ * read from the memory area.
+ *
+ * It will return zero if no valid microcode is found, and non-zero
+ * otherwise.
+ *
+ * It searches for valid microcode using intel_ucode_check_microcode()
+ * in non-strict mode.
+ *
+ * There are no alignment requirements on @bs or on the start position of
+ * the first microcode inside the memory area.  @bs and @be are pointers
+ * to a byte-aligned memory area (i.e. char* / uint8_t*).
+ *
+ * The resulting microcode area, if one is found, MIGHT BE UNALIGNED, IN
+ * WHICH CASE IT MUST BE COPIED/MOVED TO A 4-BYTE ALIGNED BUFFER BEFORE USE.
+ */
+int intel_ucode_scan_for_microcode(const void **bs, const void **be,
+				   size_t *blen, size_t *alen)
+{
+	intel_ucode_status_t r = INTEL_UCODE_INVALID_DATA;
+	const uint8_t *p, *q;
+	size_t bl, al;
+
+	if (!bs || !be || !blen || !alen)
+		return 0;
+
+	al = *alen;
+	p = *bs;
+
+	*blen = 0;
+	*be = p;
+
+	/* find first microcode */
+	while (al >= 1024) {
+		q = memchr(p, 0x01, al); /* search for a v1 header, which starts with 0x01 */
+		if (!q)
+			return 0; /* not found */
+
+		al -= (q - p);
+		p = q;
+		r = xx_intel_ucode_check_uc(p, al, 0);
+		if (r == INTEL_UCODE_NOERROR)
+			break;
+
+		if (!al)
+			return 0;
+		p++;
+		al--;
+	}
+
+	if (r != INTEL_UCODE_NOERROR)
+		return 0;
+
+	*bs = p;
+
+	/* find size of the continuous area */
+	bl = 0;
+	do {
+		const struct intel_ucode_v1_hdr * const uch = (void *)p;
+		const unsigned long int total_size = intel_ucode_v1_get_totalsize(uch);
+		p += total_size;
+		bl += total_size;
+		al -= total_size;
+
+		r = xx_intel_ucode_check_uc(p, al, 0);
+	} while (r == INTEL_UCODE_NOERROR);
+
+	*be = p;
+	*blen = bl;
+	*alen = al;
+
+	return 1;
+}
+
 /**
  * intel_ucode_foreach_signature() - run callback for every signature
  *
@@ -407,6 +555,8 @@ intel_ucode_status_t intel_ucode_check_microcode(const void * const uc, int stri
  *
  * Do NOT run this function on microcode that was not verified to be
  * correct by intel_ucode_check_microcode().
+ *
+ * @uc must be correctly aligned to a 4-byte boundary.
  */
 intel_ucode_status_t intel_ucode_foreach_signature(const void * const uc,
 					intel_ucode_sig_callback * const action,
@@ -417,7 +567,7 @@ intel_ucode_status_t intel_ucode_foreach_signature(const void * const uc,
 	unsigned int total_size, data_size;
 	uint32_t ext_table_size;
 
-	if (!action || !uc)
+	if (!action || !uc || !IS_PTR_ALIGNED(uc, sizeof(uint32_t)))
 		return INTEL_UCODE_BAD_PARAMETERS;
 
 	if (uc_header->hdrver != 1)
@@ -475,6 +625,8 @@ intel_ucode_status_t intel_ucode_foreach_signature(const void * const uc,
  *
  * A version 1 microcode bundle is a series of microcode entries, one
  * after the other, without any sort of padding.
+ *
+ * @uc_bundle must be correctly aligned to a 4-byte boundary.
  */
 intel_ucode_status_t intel_ucode_foreach_microcode(
 			const void * const uc_bundle,
@@ -486,7 +638,7 @@ intel_ucode_status_t intel_ucode_foreach_microcode(
 	size_t leftover = uc_bundle_size;
 	unsigned int uc_count;
 
-	if (!uc || !action)
+	if (!action || !uc || !IS_PTR_ALIGNED(uc, sizeof(uint32_t)))
 		return INTEL_UCODE_BAD_PARAMETERS;
 
 	/* try to guess bundle version */
