@@ -66,6 +66,7 @@ static enum { /* actions */
 	IUCODE_DO_LOADFILE = 0x0010,
 	IUCODE_DO_WRITEFWN = 0x0020,
 	IUCODE_DO_WRITEFWE = 0x0040,
+	IUCODE_DO_WRITFWNA = 0x0080,
 
 	IUCODE_F_UCSELECT  = 0x1000,
 
@@ -73,7 +74,8 @@ static enum { /* actions */
 				| IUCODE_DO_WRITEUC
 				| IUCODE_DO_WRITEFW
 				| IUCODE_DO_WRITEFWN
-				| IUCODE_DO_WRITEFWE,
+				| IUCODE_DO_WRITEFWE
+				| IUCODE_DO_WRITFWNA,
 } command_line_actions;
 
 static char *progname;
@@ -89,11 +91,13 @@ static char *write_microcode = NULL;
 static char *write_early_firmware = NULL;
 static char *write_firmware = NULL;
 static char *write_named = NULL;
+static char *write_named_all = NULL;
 
 typedef enum { /* File type */
 	INTEL_UC_FT_UNKNOWN = 0,
 	INTEL_UC_FT_DAT,
 	INTEL_UC_FT_BIN,
+	INTEL_UC_FT_SCAN,
 } intel_ucode_file_type_t;
 static intel_ucode_file_type_t ucfiletype = INTEL_UC_FT_UNKNOWN;
 
@@ -128,6 +132,8 @@ enum {
                                  *  + When INTEL_UCLE_HASXST is set,
                                  *    INTEL_UCLE_NOWR is reserved for
                                  *    uclist_annotate_extsig_dup() */
+    INTEL_UCLE_DUPSIG = 0x0200, /* Duplicate signature.  The first one
+                                 * in load order won't have this flag */
 };
 
 /* signature single-linked list */
@@ -147,7 +153,7 @@ static struct intel_uclist_entry *microcodes = NULL;
 
 /* UI helpers */
 #define UCODE_ID_MAX_LEN 22
-struct microcode_interator_data {
+struct microcode_iterator_data {
 	const struct microcode_bundle *current_bundle;
 	unsigned long int total_signature_count;
 	unsigned long int total_unique_sig_count;
@@ -155,7 +161,7 @@ struct microcode_interator_data {
 	unsigned int current_uc;
 	char current_uc_id_str[UCODE_ID_MAX_LEN];
 };
-static struct microcode_interator_data microcode_interator_data;
+static struct microcode_iterator_data microcode_iterator_data;
 
 /* Filter masks */
 struct microcode_filter_entry {
@@ -221,6 +227,7 @@ static void fix_fds(const int fd, const int fl)
 	if (nfd == -1 || dup2(nfd, fd) == -1) {
 		print_err("could not attach /dev/null to file descriptor %d: %s",
 			  fd, strerror(errno));
+		/* if (nfd != -1) close(nfd); - disabled as we're going to exit() now */
 		exit(EXIT_SWFAILURE);
 	}
 	if (nfd != fd)
@@ -330,7 +337,7 @@ static void xx_uclist_remove_shadowed(const unsigned int gid,
  */
 static int uclist_add_signature(const unsigned int id,
 			const unsigned int gid,
-			const uint32_t flags,
+			uint32_t flags,
 			const uint32_t cpuid,
 			const uint32_t pf_mask,
 			const int32_t uc_rev,
@@ -348,6 +355,8 @@ static int uclist_add_signature(const unsigned int id,
 	pnext = uclist;
 	n = NULL;
 
+	flags &= ~INTEL_UCLE_DUPSIG;
+
 	/* search linked list for first insertion point */
 	while (likely(*pnext && (*pnext)->cpuid < cpuid))
 		pnext = &((*pnext)->next);
@@ -356,6 +365,7 @@ static int uclist_add_signature(const unsigned int id,
 	while (likely(*pnext) && (*pnext)->cpuid == cpuid) {
 		if ((*pnext)->pf_mask == pf_mask && (*pnext)->uc_rev == uc_rev) {
 			res = EEXIST;
+			flags |= INTEL_UCLE_DUPSIG;
 			n = *pnext;
 		}
 		pnext = &((*pnext)->next);
@@ -422,7 +432,7 @@ static int uclist_add_signature(const unsigned int id,
  */
 static int uclist_merge_signature(const unsigned int id,
 			const unsigned int gid,
-			const uint32_t flags,
+			uint32_t flags,
 			const uint32_t cpuid,
 			const uint32_t pf_mask,
 			const int32_t uc_rev,
@@ -436,6 +446,8 @@ static int uclist_merge_signature(const unsigned int id,
 	if (unlikely(!cpuid || !uc || !uclist))
 		return EINVAL;
 
+	flags &= ~INTEL_UCLE_DUPSIG;
+
 	pnext = uclist;
 
 	/* note: pf_mask *can* be zero on old processors */
@@ -693,6 +705,11 @@ static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length,
 		err = -errno; /* negative! */
 		goto err_out;
 	}
+	if (!pos && feof(fp)) {
+		/* empty data file */
+		err = ENOENT;
+		goto err_out;
+	}
 
 	length = pos * sizeof(uint32_t);
 	if (length < 1024) {
@@ -726,7 +743,44 @@ err_out:
 	return err;
 }
 
-static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
+static int scan_and_pack_microcodes(uint8_t * const mcb, size_t * const mcb_size)
+{
+	uint8_t *mcb_e, *bs, *be;
+	size_t  mcb_len, b_len;
+	unsigned int bc;
+	int r;
+
+	mcb_len = *mcb_size;
+	mcb_e = mcb;
+	bs = mcb;
+	bc = 0;
+
+	do {
+		r = intel_ucode_scan_for_microcode((const void **)&bs,
+						   (const void **)&be,
+						   &b_len, &mcb_len);
+		if (r) {
+			print_msg(3, "microcode scan: microcode block at position %zu, size %zu bytes",
+				  bs - mcb, b_len);
+
+			if (mcb_e != bs)
+				memmove(mcb_e, bs, b_len);
+			mcb_e += b_len;
+			bs = be;
+			bc++;
+		}
+	} while (r);
+
+	mcb_len = mcb_e - mcb;
+	*mcb_size = mcb_len;
+
+	print_msg(3, "microcode scan: found %u microcode block(s), total size: %zu bytes",
+		  bc, mcb_len);
+
+	return (mcb_len) ? 0 : ENOENT;
+}
+
+static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length, int scan)
 {
 	const size_t buffer_size_granularity = IUCODE_TYPICAL_MCU_FILE_SIZE;
 	const size_t buffer_limit = IUCODE_MAX_MCU_FILE_SIZE;
@@ -748,7 +802,7 @@ static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 	/* Try to get file size beforehand */
 	if (fstat(fd, &stat))
 		return -errno; /* negative! */
-	if (S_ISREG(stat.st_mode) && stat.st_size > 0) {
+	if (S_ISREG(stat.st_mode)) {
 		if ((size_t)stat.st_size > buffer_limit)
 			return EFBIG;
 		mcb_size = stat.st_size;
@@ -760,7 +814,7 @@ static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 
 	/* sanity check size */
 	if (mcb_size < 1024) {
-		err = EINVAL;
+		err = (mcb_size) ? EINVAL : ENOENT;
 		goto err_out;
 	}
 
@@ -810,16 +864,24 @@ static int load_intel_microcode_bin(int fd, void **mcb, size_t *mcb_length)
 
 	mcb_size = pos;
 	if (mcb_size < 1024) {
-		err = EINVAL;
+		err = (mcb_size) ? EINVAL : ENOENT;
 		goto err_out;
 	}
 
+	if (scan) {
+		err = scan_and_pack_microcodes(mcb_buffer, &mcb_size);
+		if (err)
+			goto err_out;
+	}
+
 	rp = realloc(mcb_buffer, mcb_size);
-	if (!rp) {
+	if (!rp && mcb_size) {
 		err = ENOMEM;
 		goto err_out;
 	}
 
+	/* rp is allowed to be NULL when mcb_size == 0 */
+
 	*mcb = rp;
 	*mcb_length = mcb_size;
 
@@ -846,8 +908,15 @@ static int load_intel_microcode_file(int fd,
 	unsigned long int tl = 0;
 
 	switch (ftype) {
+	case INTEL_UC_FT_SCAN:
 	case INTEL_UC_FT_BIN:
-		err = load_intel_microcode_bin(fd, &mcb, &mcb_length);
+		{
+			const int scanmode = !!(ftype == INTEL_UC_FT_SCAN);
+
+			print_msg(3, "%s: loading (%sbinary mode)", fn,
+				     (scanmode) ? "microcode recovery " : "");
+			err = load_intel_microcode_bin(fd, &mcb, &mcb_length, scanmode);
+		}
 		break;
 
 	case INTEL_UC_FT_DAT:
@@ -855,6 +924,7 @@ static int load_intel_microcode_file(int fd,
 			fp = fdopen(fd, "r");
 
 		if (fp) {
+			print_msg(3, "%s: loading (.dat mode)", fn);
 			err = load_intel_microcode_dat(fp, &mcb, &mcb_length, &tl);
 		} else {
 			err = -errno;
@@ -870,7 +940,24 @@ static int load_intel_microcode_file(int fd,
 		goto err_out;
 	}
 
-	if (!err && mcb) {
+	/*
+	 * If a loader returns a zero-length microcode bundle (mcb_length
+	 * was set to 0 by the loader, or mcb is NULL), or an error, the
+	 * bundle will NOT be stored for further processing.
+	 *
+	 * If err == 0, the bundle will be assigned an ID, and listed to
+	 * the user when in verbose mode.
+	 *
+	 * If err == ENOENT, the user is notified that no microcodes were
+	 * found in that data file and that the file is being skipped.
+	 * It will not be reported as an error to the caller.
+	 *
+	 * Otherwise, the user will be notified of an error, and the error
+	 * will be reported to the caller, which may abort execution of
+	 * iucode_tool.
+	 */
+
+	if (!err && mcb && mcb_length) {
 		err = add_intel_microcode_bundle(fn, next_bundle_id,
 						 mcb, mcb_length);
 		if (!err)
@@ -883,6 +970,10 @@ static int load_intel_microcode_file(int fd,
 			  next_bundle_id, fn, mcb_length);
 		next_bundle_id++;
 		break;
+	case ENOENT:
+		print_err("%s: no microcodes found in data file, skipping...", fn);
+		err = 0;
+		break;
 	case ENOMEM:
 		print_err("%s: could not allocate memory while loading", fn);
 		break;
@@ -1011,6 +1102,11 @@ static int load_intel_microcode(const char * path,
 					err = -err;
 					continue;
 				}
+				print_msg(3, "%s: reading directory", fn);
+				/*
+				 * fd is now out-of-scope due to sucessful
+				 * fdopendir(fd), and must NOT be closed.
+				 */
 				fd = -1;
 			} else {
 				print_msg(1, "%s: skipping nested directory: %s",
@@ -1019,8 +1115,10 @@ static int load_intel_microcode(const char * path,
 			continue;
 		}
 
-		if (S_ISREG(st.st_mode) && st.st_size == 0)
+		if (S_ISREG(st.st_mode) && st.st_size == 0) {
+			print_msg(3, "skipping empty file: %s", fn);
 			continue;
+		}
 
 		ftype = baseftype;
 		if (ftype == INTEL_UC_FT_UNKNOWN && S_ISREG(st.st_mode)) {
@@ -1691,7 +1789,7 @@ static int xx_process_ucode_signature_cb(void *userdata,
 				const void * const uc,
 				unsigned int const uc_size)
 {
-	struct microcode_interator_data * const ctx = userdata;
+	struct microcode_iterator_data * const ctx = userdata;
 	struct intel_ucode_metadata m;
 	intel_ucode_status_t s;
 	int add_status;
@@ -1778,7 +1876,7 @@ static int xx_process_ucode_entry_cb(void *userdata,
 			         unsigned int const uc_count,
 			         const void * const uc)
 {
-	struct microcode_interator_data * const ctx = userdata;
+	struct microcode_iterator_data * const ctx = userdata;
 	intel_ucode_status_t s;
 
 	assert(ctx);
@@ -1788,7 +1886,7 @@ static int xx_process_ucode_entry_cb(void *userdata,
 			    ctx->current_uc_id_str,
 			    sizeof(ctx->current_uc_id_str));
 
-	s = intel_ucode_check_microcode(uc, strict_checks);
+	s = intel_ucode_check_microcode(uc, 0, strict_checks);
 	if (s != INTEL_UCODE_NOERROR) {
 		print_err("microcode %s: %s", ctx->current_uc_id_str,
 			  intel_ucode_errstr(s));
@@ -1812,7 +1910,7 @@ static int do_process_microcodes(void)
 	intel_ucode_status_t s;
 	struct microcode_bundle *mcb;
 
-	memset(&microcode_interator_data, 0, sizeof(microcode_interator_data));
+	memset(&microcode_iterator_data, 0, sizeof(microcode_iterator_data));
 	mcb = microcode_bundles;
 
 	while (mcb) {
@@ -1825,10 +1923,10 @@ static int do_process_microcodes(void)
 					mcb->id);
 			}
 		}
-		microcode_interator_data.current_bundle = mcb;
+		microcode_iterator_data.current_bundle = mcb;
 		s = intel_ucode_foreach_microcode(mcb->data, mcb->size,
 						xx_process_ucode_entry_cb,
-						&microcode_interator_data);
+						&microcode_iterator_data);
 		if (s != INTEL_UCODE_NOERROR) {
 			if (s != INTEL_UCODE_CALLBACK_ERROR)
 				print_err("microcode bundle %s: %s",
@@ -1843,9 +1941,9 @@ static int do_process_microcodes(void)
 	}
 
 	print_msg(2, "processed %lu valid microcode(s), %lu signature(s), %lu unique signature(s)",
-		  microcode_interator_data.total_entry_count,
-		  microcode_interator_data.total_signature_count,
-		  microcode_interator_data.total_unique_sig_count);
+		  microcode_iterator_data.total_entry_count,
+		  microcode_iterator_data.total_signature_count,
+		  microcode_iterator_data.total_unique_sig_count);
 
 	/*
 	 * we iterate over all selected microcodes only once, either
@@ -1921,17 +2019,19 @@ static int do_upload_microcode(const char * const filename)
 	return upload_intel_microcodes(filename, microcodes);
 }
 
-static int do_write_named(const char * const dirname)
+static int do_write_named(const char * const dirname,
+			  const struct intel_uclist_entry * const ucl)
 {
 	char fn[35]; /* "s%08X_m%08X_r%08X.fw" */
 	int dirfd;
 	struct stat st;
 
-	struct intel_uclist_entry e, *p;
+	struct intel_uclist_entry e;
+	const struct intel_uclist_entry *p;
 	unsigned long int count;
 	int rc;
 
-	if (!microcodes)
+	if (!ucl)
 		return 0;
 
 	dirfd = open(dirname, O_RDONLY);
@@ -1957,20 +2057,22 @@ static int do_write_named(const char * const dirname)
 
 	print_msg(1, "Writing microcode file(s) into %s", dirname);
 
-	p = microcodes;
+	p = ucl;
 	count = 0;
 	rc = 0;
 
 	while (p && !rc) {
-		snprintf(fn, sizeof(fn), "s%08X_m%08X_r%08X.fw",
-			     p->cpuid, p->pf_mask, p->uc_rev);
+		if (!(p->flags & INTEL_UCLE_DUPSIG)) {
+			snprintf(fn, sizeof(fn), "s%08X_m%08X_r%08X.fw",
+				     p->cpuid, p->pf_mask, p->uc_rev);
 
-		memcpy(&e, p, sizeof(e));
-		e.next = NULL;
-		e.flags &= ~INTEL_UCLE_NOWR;
-		rc = write_intel_microcodes(dirfd, fn, 0, &e);
-		if (!rc)
-			count++;
+			memcpy(&e, p, sizeof(e));
+			e.next = NULL;
+			e.flags &= ~INTEL_UCLE_NOWR;
+			rc = write_intel_microcodes(dirfd, fn, 0, &e);
+			if (!rc)
+				count++;
+		}
 
 		p = p->next;
 	}
@@ -2292,6 +2394,7 @@ enum {
 	IUCODE_ARGP_DATEFSTRICT,
 	IUCODE_ARGP_DATEFLOOSE,
 	IUCODE_ARGP_EIRFS,
+	IUCODE_ARGP_WRITENAMEDALL,
 };
 
 static struct argp_option cmdline_options[] = {
@@ -2302,8 +2405,9 @@ static struct argp_option cmdline_options[] = {
 
 	{ NULL, 't', "type", 0,
 	   "Sets input file type for the next microcode files. The type is "
-	   "a single character: \"b\" (binary), \"d\" (Intel .dat), or \"a\" "
-	   "(type will be selected by filename suffix)",
+	   "a single character: \"b\" (binary), \"d\" (Intel .dat), \"r\" "
+	   "(search inside unknown binary data), or \"a\" (type will be "
+	   "selected by filename suffix)",
 	  10 },
 
 	{ NULL, 's', "! | [!]signature[,pf_mask]", 0,
@@ -2361,15 +2465,21 @@ static struct argp_option cmdline_options[] = {
 	   "Write selected microcodes to a file in binary format.  "
 	   "The binary format is suitable to be uploaded to the kernel",
 	  40 },
-	{ "write-named-to", 'W', "directory", 0,
-	   "Write selected microcodes to files in the specified directory, "
-	   "in binary format.  The file name will reflect the microcode "
-	   "signature, mask and revision",
-	  40 },
 	{ "write-earlyfw", IUCODE_ARGP_EIRFS, "file", 0,
 	   "Write selected microcodes to an early initramfs file, which "
 	   "should be prepended to the regular initramfs",
 	  40 },
+	{ "write-named-to", 'W', "directory", 0,
+	   "Write selected microcodes to files in the specified directory, "
+	   "in binary format.  The file name will reflect the microcode "
+	   "signature, mask and revision",
+	  41 },
+	{ "write-all-named-to", IUCODE_ARGP_WRITENAMEDALL, "directory", 0,
+	   "Write every microcode to files in the specified directory, "
+	   "in binary format.  The file name will reflect the microcode "
+	   "signature, mask and revision.  This is the only way to write "
+	   "out every revision of a microcode",
+	  42 },
 
 	{ "overwrite", IUCODE_ARGP_UNLINK, NULL, 0,
 	   "Unlink (remove) destination files before writing",
@@ -2572,6 +2682,9 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 		case 'b': /* .bin */
 			ucfiletype = INTEL_UC_FT_BIN;
 			break;
+		case 'r': /* search inside binary data */
+			ucfiletype = INTEL_UC_FT_SCAN;
+			break;
 		case 'a': /* any (detect) */
 			ucfiletype = INTEL_UC_FT_UNKNOWN;
 			break;
@@ -2629,6 +2742,16 @@ static error_t cmdline_do_parse_arg(int key, char *arg,
 
 		command_line_actions |= IUCODE_DO_WRITEFWN;
 		break;
+	case IUCODE_ARGP_WRITENAMEDALL:
+		if (command_line_actions & IUCODE_DO_WRITFWNA)
+			argp_error(state,
+				   "--write-all-named-to option can be specified only once");
+
+		write_named_all = strdup(arg);
+
+		command_line_actions |= IUCODE_DO_WRITFWNA;
+		break;
+
 
 	case IUCODE_ARGP_UNLINK:
 		unlink_files = 1;
@@ -2779,7 +2902,9 @@ int main(int argc, char *argv[])
 			if (!rc && write_firmware)
 				rc = do_write_firmware(write_firmware);
 			if (!rc && write_named)
-				rc = do_write_named(write_named);
+				rc = do_write_named(write_named, microcodes);
+			if (!rc && write_named_all)
+				rc = do_write_named(write_named_all, all_microcodes);
 
 			if (rc) {
 				rc = EXIT_SWFAILURE;
